==> libaqueous-0.0.12.tar.gz: lib/src/client/room/room.dart <==
import 'package:libaqueous/src/api/events/model/event.dart';
import 'package:libaqueous/src/api/events/model/event_type.dart';
import 'package:libaqueous/src/api/events/model/tokens_chunk_events.dart';
import 'package:libaqueous/src/api/sync/model/room_sync_account_data.dart';
import 'package:libaqueous/src/api/sync/model/room_sync_state.dart';
import 'package:libaqueous/src/api/sync/model/room_sync_timeline.dart';
import 'package:libaqueous/src/client/client.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_members.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_tag.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';
import 'package:rxdart/rxdart.dart';

mixin RoomEmitter {
  final _timelineSubject = BehaviorSubject<RoomTimeline>();

  Observable<RoomTimeline> get timelineObservable => _timelineSubject.stream;

  final _stateSubject = BehaviorSubject<RoomState>();

  Observable<RoomState> get stateObservable => _stateSubject.stream;

  final _membersSubject = BehaviorSubject<RoomMembers>();

  Observable<RoomMembers> get membersObservable => _membersSubject.stream;
}

class Room with RoomEmitter {
  Room(this.client, this.state, this.accountData, this.members, this.timeline) {
    _timelineSubject.add(timeline);
    _membersSubject.add(members);
  }

  String get id => state.roomID;

  String get displayName =>
      (state.name ??
          (state.isDirect
              ? client.users[state.directUserID]?.state?.name
              : null)) ??
      state.canonicalAlias ??
      state.roomID;

  Uri get avatarUrl =>
      state.avatarUrl ??
      (state.isDirect
          ? client.users[state.directUserID]?.state?.avatarUrl
          : null);

  bool get isFavourite =>
      accountData.tags?.containsKey(RoomTag.ROOM_TAG_FAVOURITE) ?? false;

  bool get isLowPriority =>
      accountData.tags?.containsKey(RoomTag.ROOM_TAG_LOW_PRIORITY) ?? false;

  final RoomState state;
  final RoomAccountData accountData;
  final RoomTimeline timeline;
  final RoomMembers members;

  final Client client;

  Future<void> processRoomSyncState(RoomSyncState syncState,
      bool isHistory) async {
    if (syncState?.events?.isEmpty ?? true) return;

    for (final event in syncState.events) {
      await processStateEvent(event);
    }
  }

  Future<void> processStateEvent(Event event) async {
    switch (event.type) {
      case EventType.STATE_ROOM_NAME:
        state.name = event.content["name"];
        break;
      case EventType.STATE_ROOM_TOPIC:
        state.topic = event.content["topic"];
        break;
      case EventType.STATE_ROOM_AVATAR:
        state.avatarUrl = Uri.parse(event.content["url"] ?? "");
        break;
      case EventType.STATE_CANONICAL_ALIAS:
        state.canonicalAlias = event.content["alias"];
        break;
      case EventType.STATE_ROOM_MEMBER:
        await saveEventSender(event);
        break;
    }
  }

  // Why is this needed? The user might already left the room, so a full user list does not contain it.
  Future<void> saveEventSender(Event event) async {
    final avatarUrl = event.content.containsKey("avatar_url") &&
            event.content["avatar_url"] != null
        ? Uri.parse(event.content["avatar_url"])
        : null;

    // TODO: Terrible logic.
    if (!client.users.containsKey(event.sender) &&
        !await client.store.containsUserID(event.sender)) {
      final user = User(
          this.client,
          UserState()
            ..userID = event.sender
            ..name = event.content["displayname"]
            ..avatarUrl = avatarUrl);
      client.users[event.sender] = user;
      await client.store.addUserID(event.sender);
      await client.store.storeUserState(event.sender, user.state);
    } else {
      final user = await client.userOrCreate(event.sender);

      if (event.content.containsKey("avatar_url")) {
        user.state.avatarUrl = avatarUrl;
      }

      if (event.content.containsKey("displayname")) {
        user.state.name = event.content["displayname"];
      }

      await client.store.storeUserState(event.sender, user.state);
    }
  }

  Future<void> processRoomSyncTimeline(RoomSyncTimeline syncTimeline) async {
    if (syncTimeline?.events?.isEmpty ?? true) return;

    bool processedStateEvent = false;

    syncTimeline.events.first.mToken =
        syncTimeline.prevToken; // Set mToken for first event.

    if (timeline.prevToken?.isEmpty ?? true) {
      timeline.prevToken = syncTimeline.prevToken;
    }

    for (final event in syncTimeline.events) {
      event.roomID = id; // VERY IMPORTANT

      await client.loadUserIntoMemory(event.sender);

      if (EventType.isStateEvent(event.type)) {
        processedStateEvent = true;

        await processStateEvent(event);
      }
      await processTimelineEvent(event, false);
    }

    final tokensChunkEvents = TokensChunkEvents()
      ..events = syncTimeline.events
      ..start = syncTimeline.prevToken;

    await client.store.storeRoomEvents(id, tokensChunkEvents, false);

    if (processedStateEvent) {
      _stateSubject.add(state);
    }

    _timelineSubject.add(timeline);
  }

  Future<void> processBackTimelineEvents(
      TokensChunkEvents tokensChunkEvents) async {
    if (tokensChunkEvents?.events?.isEmpty ?? true) return;

    for (final event in tokensChunkEvents.events) {
      // Load user on demand.
      if (event.stateKey?.isNotEmpty ?? false) {
        await client.loadUserIntoMemory(event.stateKey);
      }
      await client.loadUserIntoMemory(event.sender);

      await processTimelineEvent(event, true);
    }
  }

  Future<void> processTimelineEvent(Event event, bool isHistory) async {
    if (isHistory) {
      timeline.events.insert(0, event);
    } else {
      timeline.events.add(event);
    }
  }

  Future<void> processRoomSyncAccountData(
      RoomSyncAccountData syncAccountData) async {
    if (syncAccountData?.events?.isEmpty ?? true) return;

    accountData.tags?.clear();

    for (final event in syncAccountData.events) {
      await processAccountDataEvent(event);
    }
  }

  Future<void> processAccountDataEvent(Event event) async {
    if (event.type == "m.tag") {
      if (event.content["tags"] is Map) {
        final tags = (event.content["tags"]) as Map<String, dynamic>;
        if (tags.isEmpty) return;

        tags.forEach((String tag, dynamic data) {
          double order = 0.0;

          if (data is Map<String, dynamic>) {
            final tagData = data as Map<String, dynamic>;
            if (tagData["order"] is int) {
              order = tagData["order"].toDouble();
            } else if (tagData["order"] is double) {
              order = tagData["order"];
            }
          }

          accountData.tags[tag] = order;
        });
      }
    }
  }

  Future<void> postMessage(Map<String, dynamic> json) async {
    await client.api.roomAPI
        .send(this.id, "m.room.message", client.transactionID, json);
  }

  Future<void> postPlainText(String text) async {
    await postMessage({"msgtype": "m.text", "body": text});
  }

  Future<void> postEmote(String text) async {
    await postMessage({"msgtype": "m.emote", "body": text});
  }

  Future<void> getEarlierMessages(int limit) async {
    print("getEarlierMessages(): prevToken: " + timeline.prevToken.toString());

    final tokensChunkEvents =
    await client.store.getEarlierMessages(id, timeline.prevToken, limit);
    if (tokensChunkEvents == null) {
      print("Fetching events from homeserver...");
      await getPreviousEvents(limit);
    } else {
      timeline.prevToken = tokensChunkEvents.end;

      await processBackTimelineEvents(tokensChunkEvents);
    }

    _timelineSubject.add(timeline);
  }

  Future<void> getPreviousEvents(int limit) async {
    final resp = (await client.api.roomAPI
            .getMessages(this.id, timeline.prevToken, "", "b", limit, ""))
        .body;
    if (resp?.chunk?.isEmpty ?? true) return;

    timeline.prevToken = resp.end;

    final tokensChunkEvents = TokensChunkEvents()
      ..events = resp.chunk
      ..start = resp.start
      ..end = resp.end;

    tokensChunkEvents.events.last.mToken = tokensChunkEvents.end;

    await client.store.storeBackToken(id, tokensChunkEvents.end);

    await client.store.storeRoomEvents(id, tokensChunkEvents, true);

    await processBackTimelineEvents(tokensChunkEvents);
  }

  Future<void> fetchMembers() async {
    print("fetchMembers(): Fetching a complete list of room members...");

    final resp = (await client.api.roomAPI.getMembers(id)).body;
    if (resp?.chunk?.isEmpty ?? true) return;

    members.hasFullMembers = true;
    members.userIDs = resp.chunk.map((e) => e.stateKey).toList();

    for (final e in resp.chunk) {
      if (client.users.containsKey(e.stateKey) ||
          await client.store.containsUserID(e.stateKey)) continue;

      final user = await client.userOrCreate(e.stateKey);
      user.state.avatarUrl = e.content["avatar_url"] == null
          ? null
          : Uri.parse(e.content["avatar_url"]);
      user.state.name = e.content["displayname"];

      await client.store.addUserID(user.id);
      await client.store.storeUserState(user.id, user.state);
    }

    await client.store.storeRoomMembers(id, members);

    print("fetchMembers(): Got " +
        members.userIDs.length.toString() +
        " members");

    _membersSubject.add(members);
  }
}

==> libaqueous-0.0.13.tar.gz: lib/src/client/room/room.dart <==
import 'package:libaqueous/src/api/events/model/event.dart';
import 'package:libaqueous/src/api/events/model/event_type.dart';
import 'package:libaqueous/src/api/events/model/tokens_chunk_events.dart';
import 'package:libaqueous/src/api/sync/model/room_sync_account_data.dart';
import 'package:libaqueous/src/api/sync/model/room_sync_state.dart';
import 'package:libaqueous/src/api/sync/model/room_sync_timeline.dart';
import 'package:libaqueous/src/client/client.dart';
import 'package:libaqueous/src/client/room/room_account_data.dart';
import 'package:libaqueous/src/client/room/room_members.dart';
import 'package:libaqueous/src/client/room/room_state.dart';
import 'package:libaqueous/src/client/room/room_tag.dart';
import 'package:libaqueous/src/client/room/room_timeline.dart';
import 'package:libaqueous/src/client/user/user.dart';
import 'package:libaqueous/src/client/user/user_state.dart';
import 'package:rxdart/rxdart.dart';

mixin RoomEmitter {
  final _timelineSubject = BehaviorSubject<RoomTimeline>();

  Observable<RoomTimeline> get timelineObservable => _timelineSubject.stream;

  final _stateSubject = BehaviorSubject<RoomState>();

  Observable<RoomState> get stateObservable => _stateSubject.stream;

  final _membersSubject = BehaviorSubject<RoomMembers>();

  Observable<RoomMembers> get membersObservable => _membersSubject.stream;
}

class Room with RoomEmitter {
  Room(this.client, this.state, this.accountData, this.members, this.timeline) {
    _timelineSubject.add(timeline);
    _membersSubject.add(members);
  }

  String get id => state.roomID;

  String get displayName =>
      (state.name ??
          (state.isDirect
              ? client.users[state.directUserID]?.state?.name
              : null)) ??
      state.canonicalAlias ??
      state.roomID;

  Uri get avatarUrl =>
      state.avatarUrl ??
      (state.isDirect
          ? client.users[state.directUserID]?.state?.avatarUrl
          : null);

  bool get isFavourite =>
      accountData.tags?.containsKey(RoomTag.ROOM_TAG_FAVOURITE) ?? false;

  bool get isLowPriority =>
      accountData.tags?.containsKey(RoomTag.ROOM_TAG_LOW_PRIORITY) ?? false;

  final RoomState state;
  final RoomAccountData accountData;
  final RoomTimeline timeline;
  final RoomMembers members;

  final Client client;

  Future<void> processRoomSyncState(RoomSyncState syncState,
      bool isHistory) async {
    if (syncState?.events?.isEmpty ?? true) return;

    for (final event in syncState.events) {
      await processStateEvent(event);
    }
  }

  Future<void> processStateEvent(Event event) async {
    switch (event.type) {
      case EventType.STATE_ROOM_NAME:
        state.name = event.content["name"];
        break;
      case EventType.STATE_ROOM_TOPIC:
        state.topic = event.content["topic"];
        break;
      case EventType.STATE_ROOM_AVATAR:
        state.avatarUrl = Uri.parse(event.content["url"] ?? "");
        break;
      case EventType.STATE_CANONICAL_ALIAS:
        state.canonicalAlias = event.content["alias"];
        break;
      case EventType.STATE_ROOM_MEMBER:
        await saveEventSender(event);
        break;
    }
  }

  // Why is this needed? The user might already left the room, so a full user list does not contain it.
  Future<void> saveEventSender(Event event) async {
    final avatarUrl = event.content.containsKey("avatar_url") &&
            event.content["avatar_url"] != null
        ? Uri.parse(event.content["avatar_url"])
        : null;

    // TODO: Terrible logic.
    if (!client.users.containsKey(event.sender) &&
        !await client.store.containsUserID(event.sender)) {
      final user = User(
          this.client,
          UserState()
            ..userID = event.sender
            ..name = event.content["displayname"]
            ..avatarUrl = avatarUrl);
      client.users[event.sender] = user;
      await client.store.addUserID(event.sender);
      await client.store.storeUserState(event.sender, user.state);
    } else {
      final user = await client.userOrCreate(event.sender);

      if (event.content.containsKey("avatar_url")) {
        user.state.avatarUrl = avatarUrl;
      }

      if (event.content.containsKey("displayname")) {
        user.state.name = event.content["displayname"];
      }

      await client.store.storeUserState(event.sender, user.state);
    }
  }

  Future<void> processRoomSyncTimeline(RoomSyncTimeline syncTimeline) async {
    if (syncTimeline?.events?.isEmpty ?? true) return;

    bool processedStateEvent = false;

    syncTimeline.events.first.mToken =
        syncTimeline.prevToken; // Set mToken for first event.

    if (timeline.prevToken?.isEmpty ?? true) {
      timeline.prevToken = syncTimeline.prevToken;
    }

    for (final event in syncTimeline.events) {
      event.roomID = id; // VERY IMPORTANT

      await client.loadUserIntoMemory(event.sender);

      if (EventType.isStateEvent(event.type)) {
        processedStateEvent = true;

        await processStateEvent(event);
      }
      await processTimelineEvent(event, false);
    }

    final tokensChunkEvents = TokensChunkEvents()
      ..events = syncTimeline.events
      ..start = syncTimeline.prevToken;

    await client.store.storeRoomEvents(id, tokensChunkEvents, false);

    if (processedStateEvent) {
      _stateSubject.add(state);
    }

    _timelineSubject.add(timeline);
  }

  Future<void> processBackTimelineEvents(
      TokensChunkEvents tokensChunkEvents) async {
    if (tokensChunkEvents?.events?.isEmpty ?? true) return;

    for (final event in tokensChunkEvents.events) {
      // Load user on demand.
      if (event.stateKey?.isNotEmpty ?? false) {
        await client.loadUserIntoMemory(event.stateKey);
      }
      await client.loadUserIntoMemory(event.sender);

      await processTimelineEvent(event, true);
    }
  }

  Future<void> processTimelineEvent(Event event, bool isHistory) async {
    if (isHistory) {
      timeline.events.insert(0, event);
    } else {
      timeline.events.add(event);
    }
  }

  Future<void> processRoomSyncAccountData(
      RoomSyncAccountData syncAccountData) async {
    if (syncAccountData?.events?.isEmpty ?? true) return;

    accountData.tags?.clear();

    for (final event in syncAccountData.events) {
      await processAccountDataEvent(event);
    }
  }

  Future<void> processAccountDataEvent(Event event) async {
    if (event.type == "m.tag") {
      if (event.content["tags"] is Map) {
        final tags = (event.content["tags"]) as Map<String, dynamic>;
        if (tags.isEmpty) return;

        tags.forEach((String tag, dynamic data) {
          double order = 0.0;

          if (data is Map<String, dynamic>) {
            final tagData = data as Map<String, dynamic>;
            if (tagData["order"] is int) {
              order = tagData["order"].toDouble();
            } else if (tagData["order"] is double) {
              order = tagData["order"];
            }
          }

          accountData.tags[tag] = order;
        });
      }
    }
  }

  Future<void> postMessage(Map<String, dynamic> json) async {
    await client.api.roomAPI
        .send(this.id, "m.room.message", client.transactionID, json);
  }

  Future<void> postPlainText(String text) async {
    await postMessage({"msgtype": "m.text", "body": text});
  }

  Future<void> postEmote(String text) async {
    await postMessage({"msgtype": "m.emote", "body": text});
  }

  Future<void> getEarlierMessages(int limit) async {
    print("getEarlierMessages(): prevToken: " + timeline.prevToken.toString());

    final tokensChunkEvents =
    await client.store.getEarlierMessages(id, timeline.prevToken, limit);
    if (tokensChunkEvents == null) {
      print("Fetching events from homeserver...");
      await getPreviousEvents(limit);
    } else {
      timeline.prevToken = tokensChunkEvents.end;

      await processBackTimelineEvents(tokensChunkEvents);
    }

    _timelineSubject.add(timeline);
  }

  Future<void> getPreviousEvents(int limit) async {
    final resp = (await client.api.roomAPI
            .getMessages(this.id, timeline.prevToken, "", "b", limit, ""))
        .body;
    if (resp?.chunk?.isEmpty ?? true) return;

    timeline.prevToken = resp.end;

    final tokensChunkEvents = TokensChunkEvents()
      ..events = resp.chunk
      ..start = resp.start
      ..end = resp.end;

    tokensChunkEvents.events.last.mToken = tokensChunkEvents.end;

    await client.store.storeBackToken(id, tokensChunkEvents.end);

    await client.store.storeRoomEvents(id, tokensChunkEvents, true);

    await processBackTimelineEvents(tokensChunkEvents);
  }

  Future<void> fetchMembers() async {
    print("fetchMembers(): Fetching a complete list of room members...");

    final resp = (await client.api.roomAPI.getMembers(id)).body;
    if (resp?.chunk?.isEmpty ?? true) return;

    members.hasFullMembers = true;
    members.userIDs = resp.chunk.map((e) => e.stateKey).toList();

    Map<String, User> userMap = {};
    for (final e in resp.chunk) {
      if (client.users.containsKey(e.stateKey) ||
          await client.store.containsUserID(e.stateKey)) continue;

      final user = User(client, UserState()
        ..userID = e.stateKey);
      user.state.avatarUrl = e.content["avatar_url"] == null
          ? null
          : Uri.parse(e.content["avatar_url"]);
      user.state.name = e.content["displayname"];

      userMap[e.stateKey] = user;
    }

    await client.store.storeRoomMembers(id, members);

    if (userMap.isNotEmpty) {
      await client.store.addUserIDBulk(userMap.keys.toList());
      await client.store.storeUserStateBulk(
          userMap.map((id, user) => MapEntry(id, user.state)));
    }

    print("fetchMembers(): Got " +
        members.userIDs.length.toString() +
        " members");

    _membersSubject.add(members);
  }
}
